'''
Gauged - https://github.com/chriso/gauged
Copyright 2014 (c) Chris O'Hara <cohara87@gmail.com>
'''

import unittest

class TestCase(unittest.TestCase):
    '''A base class for all Gauged tests'''
