'''
Gauged
https://github.com/chriso/gauged (MIT Licensed)
Copyright 2014 (c) Chris O'Hara <cohara87@gmail.com>
'''

__version__ = '0.4.0'

__version_info__ = tuple([ int(v) for v in __version__.split('.') ])
